# coding: utf-8

import ConfigParser
import os
import re
import shlex
import subprocess

import pkg_resources

from tionix_client.cmd.configure import process_dir, ConfigureException

from tionix_scheduler.core import BASE_DIR
from tionix_scheduler.core.i18n import ugettext as _


CONF_DIR = os.path.join(BASE_DIR, 'conf')

apache_config_template = u"""
Listen 10001

<VirtualHost *:10001>

    WSGIDaemonProcess tionix_scheduler \
user=tionix group=tionix processes=3 threads=10
    WSGIScriptAlias / {scheduler_module}/tionix_scheduler/wsgi.py
    WSGIProcessGroup tionix_scheduler

    ErrorLog {log_path}/tionix-scheduler.log
    CustomLog {log_path}/tionix-scheduler-access.log combined

    <Directory {scheduler_module}/tionix_scheduler>
        <Files wsgi.py>
            {auth_directive}
        </Files>
    </Directory>

    Alias /static/admin {django_module}/django/contrib/admin/static/admin
    <Directory {django_module}/django/contrib/admin/static/admin>
        {auth_directive}
    </Directory>

    Alias /static {rest_module}/rest_framework/static
    <Directory {rest_module}/rest_framework/static>
        {auth_directive}
    </Directory>

</VirtualHost>
"""


def get_existing_path(paths):
    return next((i for i in paths if os.path.exists(i)), None)


def get_module_path(module_name):
    """
    Try to find out the module dist location
    For tionix modules the best way to find location for parent package
    is to locate the location for one of the required package
    :param module_name: the module name to determine location
    :return: location for package
    """
    module_name = module_name.replace('-', '_')

    path = u'/usr/local/lib/python2.7/dist-packages'
    try:
        module = pkg_resources.get_distribution(module_name)
    except (pkg_resources.DistributionNotFound, pkg_resources.VersionConflict,
            pkg_resources.ContextualVersionConflict):
        pass

    else:
        path = module.location

    return path


def get_apache_log_path():
    paths = (
        '/var/log/apache2',
        '/var/log/apache',
        '/var/log/httpd2',
        '/var/log/httpd',
    )
    return get_existing_path(paths) or paths[0]


def get_apache_version():
    version_pattern = re.compile(r'Apache/([0-9.]*)\s')
    commands = (
        'apache2 -v',
        'httpd2 -v',
        'httpd -v',
        '/usr/sbin/apache2 -v',
        '/usr/sbin/httpd2 -v',
        '/usr/sbin/httpd -v',
    )

    version = None
    for command in commands:
        try:
            output = subprocess.check_output(shlex.split(command))
        except (subprocess.CalledProcessError, OSError):
            pass
        else:
            parsed = version_pattern.findall(output)
            if parsed:
                version = parsed[0]
                break
    return version


def get_auth_directive():
    version = get_apache_version()
    default_version = pkg_resources.parse_version(u'2.4')
    permission_24 = u'Require all granted'
    permission_22 = u'Allow from all'

    permission = permission_24
    if version:
        try:
            version = pkg_resources.parse_version(version)
        except pkg_resources.RequirementParseError:
            pass
        else:
            if version < default_version:
                permission = permission_22
    return permission


def get_config_available_path():
    paths = (
        '/etc/apache2/conf-available',
        '/etc/httpd2/conf/sites-available',
        '/etc/httpd/conf/sites-available',
    )
    return get_existing_path(paths) or paths[0]


def generate_apache_config():
    config_opts = {
        'scheduler_module': get_module_path('tionix_scheduler_client'),
        'django_module': get_module_path('django'),
        'rest_module': get_module_path('djangorestframework'),
        'log_path': get_apache_log_path(),
        'auth_directive': get_auth_directive(),
    }

    return apache_config_template.format(**config_opts)


def create_apache_config():
    apache_config_name = u'tionix-scheduler-dash.conf'
    config_path = get_config_available_path()
    config = generate_apache_config()
    if not os.path.exists(config_path):
        try:
            os.makedirs(config_path)
        except EnvironmentError:
            raise ConfigureException(_(
                u'Unable to create Apache config directory: {0}').format(
                    config_path))

    elif os.access(config_path, os.W_OK):
        with open(os.path.join(config_path, apache_config_name), 'w') as fh:
            fh.write(config)
    else:
        raise ConfigureException(_(
            u'Unable to create Apache configuration file '
            u'in directory: {0}').format(config_path))


def update_supervisor_config():
    paths = (
        u'/etc/supervisor/supervisord.conf',
        u'/etc/supervisord.conf',
    )
    conf_files = u'/etc/supervisor/conf.d/*.conf'
    config_section = u'include'
    config_key = u'files'

    config_path = get_existing_path(paths)
    if not config_path:
        msg = _(u'Unable to create sample supervisor configuration file: {0}')
        config_path = paths[0]
        init_command = u'echo_supervisord_conf > {0}'.format(config_path)
        try:
            popen = subprocess.Popen(init_command, shell=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)
            output = popen.communicate()[0]

            if popen.returncode != 0:
                raise ConfigureException(msg.format(unicode(output)))

        except OSError as err:
            raise ConfigureException(msg.format(unicode(err)))

    config = ConfigParser.ConfigParser()
    config.read(config_path)
    if not config.has_section(config_section):
        config.add_section(config_section)
    try:
        value = config.get(config_section, config_key)
    except ConfigParser.NoOptionError:
        config.set(config_section, config_key, conf_files)
    else:
        if conf_files not in value:
            config.set(
                config_section, config_key, u' '.join([value, conf_files]))
    try:
        with open(config_path, 'w') as fh:
            config.write(fh)
    except EnvironmentError as err:
        raise ConfigureException(_(
            u'Unable to update supervisor configuration file: {0}').format(
                unicode(err)))


def configure(verbose=False):
    try:
        process_dir(CONF_DIR)
        create_apache_config()
        update_supervisor_config()
    except ConfigureException as err:
        result = os.EX_SOFTWARE, err.message
    except Exception as ex:
        if verbose:
            msg = _(u'Error while configuring module: {0}').format(unicode(ex))
        else:
            msg = _(u'Unknown error. Please contact your administrator.')
        result = os.EX_SOFTWARE, msg
    else:
        result = os.EX_OK, u''

    return result
