# coding: utf-8

from __future__ import absolute_import

import datetime
import json
import logging
import traceback

import amqp

from celery.beat import SchedulingError
from celery.schedules import schedule as interval_schedule

from django.utils.dateparse import parse_datetime

from djcelery.schedulers import DatabaseScheduler, ModelEntry

from kombu import BrokerConnection
from kombu.utils.compat import timedelta_seconds

from tionix_scheduler.settings import BROKER_URL
from tionix_scheduler.timetable.models import TnxPeriodicTask

logger = logging.getLogger(__name__)


class TaskMutexError(Exception):
    pass


class TnxModelEntry(ModelEntry):

    def __init__(self, model):
        self._last_run_at = model.last_run_at
        super(TnxModelEntry, self).__init__(model)

    def is_due(self):

        is_due, next_time_to_run = super(TnxModelEntry, self).is_due()

        # interval task first run bugfix when last_run_at definitely is None
        # pylint: disable=unidiomatic-typecheck
        if type(self.schedule) is interval_schedule:
            if not self._last_run_at:
                is_due = True

        begins = self.kwargs.get('begins')
        if is_due and begins:
            try:
                begins = parse_datetime(begins)
            except (ValueError, TypeError):
                pass
            else:
                now = datetime.datetime.now(begins.tzinfo)
                if now < begins:
                    after = timedelta_seconds(begins - now)
                    is_due = False
                    next_time_to_run = after
                else:
                    self._last_run_at = now

        return is_due, next_time_to_run

    @classmethod
    def from_entry(cls, name, skip_fields=('relative', 'options'), **fields):
        options = fields.get('options') or {}
        for skip_field in skip_fields:
            fields.pop(skip_field, None)
        schedule = fields.pop('schedule')
        model_schedule, model_field = cls.to_model_schedule(schedule)
        fields[model_field] = model_schedule
        fields['args'] = json.dumps(fields.get('args') or [])
        fields['kwargs'] = json.dumps(fields.get('kwargs') or {})
        fields['queue'] = options.get('queue')
        fields['exchange'] = options.get('exchange')
        fields['routing_key'] = options.get('routing_key')
        # pylint: disable=no-member
        default_manager = TnxPeriodicTask._default_manager
        obj = default_manager.update_or_create(name=name, defaults=fields)
        if isinstance(obj, tuple):
            obj, _created = obj
        return cls(obj)

    def load_model(self):
        obj = self.model._default_manager.get(pk=self.model.pk)
        return obj


class TnxDatabaseScheduler(DatabaseScheduler):

    Entry = TnxModelEntry
    Model = TnxPeriodicTask

    def __init__(self, *args, **kwargs):
        super(TnxDatabaseScheduler, self).__init__(*args, **kwargs)
        # enable db sync after each sent task
        self.sync_every_tasks = True

    def maybe_due(self, entry, publisher=None):
        is_due, next_time_to_run = entry.is_due()

        if is_due:
            logger.info('Scheduler: Sending due task %s (%s)', entry.name,
                        entry.task)
            try:
                result = self.apply_async(entry, publisher=publisher)
            except TaskMutexError as exc:
                logger.info(unicode(exc))
            except Exception as exc:
                logger.error('Message Error: %s\n%s', exc,
                             traceback.format_stack(), exc_info=True)
            else:
                logger.debug('%s sent. id->%s', entry.task, result.id)
        return next_time_to_run

    def apply_async(self, entry, publisher=None, **kwargs):

        result = None
        entry = self.reserve(entry)
        task = self.app.tasks.get(entry.task)
        task_id = entry.name
        run_count = entry.total_run_count

        connection = BrokerConnection(BROKER_URL)

        try:
            queue_name = 'celery_beat_{0}_{1}.mutex'.format(task_id, run_count)
            queue = connection.SimpleQueue(queue_name,
                                           queue_opts={'exclusive': True})
        except amqp.exceptions.ChannelError:
            raise TaskMutexError('Task with id={0} is locked by another '
                                 'beat.'.format(task_id))
        with queue:
            task_obj = entry.load_model()
            if task_obj.total_run_count >= run_count:
                raise TaskMutexError('Task with id={0} has already been '
                                     'processed in {1} iteration.'
                                     ''.format(task_id, run_count))
            else:
                try:
                    if task:
                        result = task.apply_async(entry.args, entry.kwargs,
                                                  publisher=publisher,
                                                  **entry.options)
                    else:
                        result = self.send_task(entry.task, entry.args,
                                                entry.kwargs,
                                                publisher=publisher,
                                                **entry.options)
                except Exception as exc:
                    raise SchedulingError(
                        "Couldn't apply scheduled task {0.name}: {exc}"
                        "".format(entry, exc=exc))
                finally:
                    self._tasks_since_sync += 1
                    if self.should_sync():
                        self._do_sync()
        return result
