# coding: utf-8
# pylint: disable=abstract-method

from __future__ import absolute_import

import logging
import uuid

import importlib

from celery.app.task import Task

from django.conf import settings
from django.utils import timezone

from djcelery.models import IntervalSchedule, CrontabSchedule

from pkg_resources import iter_entry_points

from tionix_scheduler.celery import app as celery_app
from tionix_scheduler.timetable.models import TimeTable, TnxPeriodicTask

logger = logging.getLogger(__name__)


class TnxTask(Task):

    abstract = True
    _expired = False

    def __call__(self, *args, **kwargs):
        time_table = self.get_timetable(*args, **kwargs)
        if time_table:
            task = time_table.periodic_task
            if task.expires and task.expires <= timezone.now():
                self._expired = True
                return

        return super(TnxTask, self).__call__(*args, **kwargs)

    @staticmethod
    def parse_task_result(result):
        """
        Parse retval - task return value - expected tuple or list
        with text message and success bool flag.

        :param result:
        :return:
        """
        message = result
        if isinstance(result, (list, tuple)) and len(result) > 1:
            message = result[0]
            result = result[1]
            last_run_result = (TimeTable.TASK_SUCCESS if result
                               else TimeTable.TASK_FAILURE)
        else:
            last_run_result = TimeTable.TASK_UNKNOWN

        return message, last_run_result

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        """Process executed task"""
        time_table = self.get_timetable(*args, **kwargs)
        if time_table:
            # disable expired task
            if self._expired:
                self.disable_time_table(time_table)
            else:
                # disable periodic task if run count limit is reached
                # increment run_count, periodic task is not updated yet
                run_count = time_table.periodic_task.total_run_count
                run_limit = time_table.run_count_limit
                if run_limit and run_count >= run_limit:
                    self.disable_time_table(time_table)

                #  disable single task
                if not time_table.repeat:
                    self.disable_time_table(time_table)

                # Saves unique run task id.
                # Task id is needed to get run result
                time_table.last_run_task_id = task_id

                message, result = self.parse_task_result(retval)
                time_table.last_run_message = message
                time_table.last_run_result = result

                time_table.save()

    # pylint: disable=no-member
    @staticmethod
    def get_timetable(*args, **kwargs):
        time_table = None
        periodic_task_uuid_name = kwargs.get('periodic_task_uuid_name')
        if periodic_task_uuid_name:
            try:
                time_table = TimeTable.objects.get(
                    periodic_task__name=periodic_task_uuid_name)
            except TimeTable.DoesNotExist:
                pass
        return time_table

    @staticmethod
    def disable_time_table(time_table):
        time_table.periodic_task.enabled = False
        time_table.periodic_task.save()
        logger.debug(u'Task {0} was successfully disabled'.format(
            time_table.description))


class TaskItem(object):
    """
    Class describes task that becomes a celery-task.
    """

    def __init__(self, subject_type, subject_groups, code, locale_dict, task):
        super(TaskItem, self).__init__()
        # type of task subject
        self.subject_type = subject_type
        # group of task subject
        self.subject_groups = subject_groups
        # unique name of task-module
        self.unique_code = code
        # localized human-readable name of task
        self.locale_dict = locale_dict
        # current ref of task
        self.task = task

    def apply_schedule(self, schedule, *args, **kwargs):
        """
        Run task according to giving schedule.

        :param schedule: an object contains periodic description
        :type schedule: IntervalSchedule/CrontabSchedule
        """
        if not isinstance(schedule, (IntervalSchedule, CrontabSchedule)):
            raise TypeError(u'Incorrect type of "schedule". Must be '
                            u'one of (IntervalSchedule, CrontabSchedule)')

        schedule_key = {
            'IntervalSchedule': {'interval': schedule},
            'CrontabSchedule': {'crontab': schedule}
        }

        uuid_name = uuid.uuid4()

        kwargs['periodic_task_uuid_name'] = str(uuid_name)

        task = {
            'name': uuid_name,
            'task': self.task,
            'args': args,
            'kwargs': kwargs
        }

        task.update(schedule_key.get(schedule._meta.object_name))

        # Extra available params see in djcelery.models.PeriodicTask.
        task.update(*args)

        return TnxPeriodicTask.objects.create(**task)

    def get_verbose_name(self, locale=None):
        return self.locale_dict.get(locale or settings.TASK_LOCALE, '')

    verbose_name = property(get_verbose_name)


class TaskController(object):
    """
    Class controls system-wide tasks.
    """
    task_collection = {}

    @classmethod
    def populate_tasks(cls, force=False):
        if cls.task_collection and not force:
            return

        import_modules = []
        cls.task_collection = {}

        # for each entry group collect tasks
        for group in settings.ENTRY_GROUPS:
            cls.task_collection[group] = {}

            for task_entry_point in iter_entry_points(group=group):
                try:
                    func = task_entry_point.resolve()
                except Exception as err:
                    logger.error(
                        u'Skipped task "{0}" because of '
                        u'the following error: {1}'.format(
                            task_entry_point.module_name, unicode(err))
                    )
                    continue

                if not isinstance(func, Task):
                    # make task from function if it is not yet a task
                    try:
                        new_task = celery_app.task(func, base=TnxTask)
                    except Exception as err:
                        raise err
                else:
                    new_task = func

                import_modules.append(task_entry_point.module_name)

                # keep to quick access
                cls.__add_task_to_collection(group, task_entry_point, new_task)
        # celery_app.autodiscover_tasks(import_modules)
        settings.CELERY_IMPORTS = import_modules

    @classmethod
    def __add_task_to_collection(cls, entry_point, task_entry_point, task):
        task_module = importlib.import_module(task_entry_point.module_name)

        unique_code = getattr(task_module, 'unique_code')
        if unique_code is None:
            raise ImportError('Unique_code is empty. Can not import task.')

        locale = getattr(task_module, 'locale', None)

        if not isinstance(locale, dict):
            raise ImportError(
                'locale is empty. Can not import task {}.'.format(
                    task_entry_point.module_name)
            )

        cls.task_collection[entry_point].update({
            unique_code: TaskItem(
                getattr(task_module, 'subject_type', 'unknown'),
                getattr(task_module, 'subject_groups', 'unknown'),
                unique_code,
                locale,
                task.name,
            )
        })

    @classmethod
    def get_task(cls, unique_code):
        """
        Find task by unique_code defined in module.
        """
        if not cls.task_collection:
            cls.populate_tasks()
        task = None
        for _group, items in cls.task_collection.iteritems():
            for code, item in items.iteritems():
                if code == unique_code:
                    task = item
                    break
            if task is not None:
                break

        return task

TaskController.populate_tasks()
