# coding: utf-8

import os

from mock import Mock, patch

from rest_framework.test import APITestCase, APIClient

from tionix_scheduler_client.enums import ObjectTypeEnum, ObjectGroupEnum
from tionix_scheduler.tasks_controller import TaskController, TaskItem


class FakeKeystoneTokenMiddleware(object):
    def process_request(self, request):
        pass


class TimeTableTestManager(object):
    """
    Test client manager.
    """
    root = '/'
    basename = 'timetable/'
    test_unique_code = 'test_unique_code'

    def __init__(self, client):
        super(TimeTableTestManager, self).__init__()

        if client is None:
            client = APIClient()

        self.client = client
        self.test_subject_name = 'Subject test name'

        self.url = os.path.join(self.root, self.basename)

    def make_url(self, object_id):
        return '{}{}/'.format(self.url, object_id)

    def get_interval_data(self):
        """
        Returns data to create timetable and periodic task records
        with interval schedule.

        :return Timetable, Periodic Task and data to pass into request.
        :rtype dict
        """
        timetable = {
            'author_name': 'admin',
            'tenant_id': '123456789',
            'description': 'Test Interval Task',
            'unique_name': self.test_unique_code,
            'repeat': True,
            'subject': self.test_subject_name,
            'subject_type': ObjectTypeEnum.instance,
            'subject_group': ObjectGroupEnum.admin,
            'subject_id': '1',
            'begin_date': None,
            'begin_time': None,
            'timezone': "",
            'run_count_limit': None,
        }

        subject = timetable['subject']
        subject_id = timetable['subject_id']
        args = '[1,2,3,"{0}","{1}", "test"]'.format(subject_id, subject)

        periodic_task = {
            'activity': 'true',
            'interval': {
                'every': '20',
                'period': 'minutes'
            },
            'arguments': args
        }

        data = {}
        data.update(timetable)
        data.update(periodic_task)

        return timetable, periodic_task, data

    def get_crontab_data(self):
        """
        Returns data to create timetable and periodic task records
        with crontab schedule.

        :return Timetable, Periodic Task and data to pass into request.
        :rtype dict
        """
        timetable = {
            'author_name': 'admin',
            'tenant_id': '123456789',
            'description': 'Test Crontab Task',
            'unique_name': self.test_unique_code,
            'repeat': True,
            'subject': self.test_subject_name,
            'subject_type': ObjectTypeEnum.hypervisor,
            'subject_group': ObjectGroupEnum.admin,
            'subject_id': '2',
            'begin_date': None,
            'begin_time': None,
            'timezone': "",
            'run_count_limit': None,
        }
        subject = timetable['subject']
        subject_id = timetable['subject_id']
        args = '[1,2,3,"{0}","{1}", "test"]'.format(subject_id, subject)

        periodic_task = {
            'activity': 'true',
            'expires': '2018-01-01 23:59:59',
            'crontab': {
                'minute': '*',
                'hour': '*',
                'day_of_week': '*',
                'day_of_month': '01',
                'month_of_year': '01'
            },
            'arguments': args
        }

        data = {}
        data.update(timetable)
        data.update(periodic_task)

        return timetable, periodic_task, data

    def create_timetable(self, data):
        """
        Creates timetable.

        :param data: Required data of timetable to create.
        :type data: dict

        :return Response after POST request.
        :rtype django.http.response.HttpResponse
        """
        return self.client.post(self.url, data, format='json')

    def patch_timetable(self, data, object_id):
        """
        Patches timetable by id.

        :param data: Required data of timetable to patch.
        :type data: dict

        :param object_id: ID of timetable to patch.
        :type object_id: str

        :return Response after PATCH request.
        :rtype django.http.response.HttpResponse
        """
        url = self.make_url(object_id)
        return self.client.patch(url, data, format='json')

    def put_timetable(self, data, object_id):
        """
        Updates timetable by id.

        :param data: Required data of timetable to put.
        :type data: dict

        :param object_id: ID of timetable to put.
        :type object_id: str

        :return Response after PUT request.
        :rtype django.http.response.HttpResponse
        """
        url = self.make_url(object_id)
        return self.client.put(url, data, format='json')

    def delete_timetable(self, object_id):
        """
        Deleted timetable by id.

        :param object_id: ID of timetable to delete.
        :type object_id: str

        :return Response after DELETE request.
        :rtype django.http.response.HttpResponse
        """
        url = self.make_url(object_id)
        return self.client.delete(url)

    def delete_subject_timetable(self, subject_id=None, subject_type=None):
        """
        Deleted timetable by subject_id and subject_type.

        :param subject_id: ID of subject in timetable.
        :type subject_id: str

        :param subject_type: type of subject in timetable.
        :type subject_type: str

        :return Response after DELETE request.
        :rtype django.http.response.HttpResponse
        """
        data = {
            'subject_id': subject_id,
            'subject_type': subject_type,
        }
        url = os.path.join(self.url, 'tasks_monitoring/')
        return self.client.delete(url, data, format='json')

    def get_timetable(self, object_id):
        """
        Returns timetable by id.

        :param object_id: ID of timetable to get.
        :type object_id: str

        :return Response after GET request.
        :rtype django.http.response.HttpResponse
        """
        url = self.make_url(object_id)
        return self.client.get(url)

    def get_tasks_list(self, **params):
        """
        Returns available tasks to create timetable.

        :return Response after GET request.
        :rtype django.http.response.HttpResponse
        """
        url = os.path.join(self.url, 'tasks_list/')
        # put session token to request
        return self.client.get(url, data=params)

    def get_timetable_list(self):
        """
        Returns timetables.

        :return Response after GET request.
        :rtype django.http.response.HttpResponse
        """
        return self.client.get(self.url)

    def get_monitoring_tasks_list(self, params=None):
        """
        Returns periodic tasks in execution.

        :return Response after GET request.
        :rtype django.http.response.HttpResponse
        """
        url = os.path.join(self.url, 'tasks_monitoring/')
        return self.client.get(url, params)

    def get_tasks_objects(self):
        """
        Returns available object types to create timetable.
        """
        url = os.path.join(self.url, 'task_objects/')
        return self.client.get(url)


class BaseAPITestCase(APITestCase):
    test_ep_name = 'test_entry_point'

    @classmethod
    def setUpClass(cls):
        super(BaseAPITestCase, cls).setUpClass()

        cls.test_task_items = [
            TaskItem(
                'test_type',
                ['test_group'],
                'test_unique_code',
                {'en': u'Test task 1', 'ru': u'Тестовое задание 1'},
                'test_tionix_tasks.task.test',
            ),
            TaskItem(
                'test_type',
                ['test_group'],
                'test_unique_code_2',
                {'en': u'Test task 2', 'ru': u'Тестовое задание 2'},
                'test_tionix_tasks.task.test',
            ),
            TaskItem(
                'test_type_2',
                ['test_group_2'],
                'test_unique_code_3',
                {'en': u'Test task 3', 'ru': u'Тестовое задание 3'},
                'test_tionix_tasks.task.test',
            )
        ]

        TaskController.task_collection.update({
            BaseAPITestCase.test_ep_name: {
                task.unique_code: task for task in cls.test_task_items}})

    def _get_user_info_patch(self, username, tenants=None):
        if tenants is None:
            tenants = [self.test_tenant_id]
        patch('tionix_scheduler.timetable.views.get_user_info',
              new=Mock(return_value=(username, tenants))).start()

    def setUp(self):

        self.manager = TimeTableTestManager(self.client)
        patch('tionix_scheduler.timetable.serializers.get_keystone_client',
              new=Mock(return_value=True)).start()
        patch('tionix_scheduler.middleware.KeystoneTokenMiddleware',
              new=FakeKeystoneTokenMiddleware).start()

        self.test_tenant_id = '123456789'
        self.test_username = 'admin'

        self._get_user_info_patch(self.test_username)
