# coding: utf-8
# pylint: disable=no-member

import datetime
import json

import dateutil.parser

import mock

from django.conf import settings

from rest_framework import status
from rest_framework.test import APIClient, APITestCase

from tionix_scheduler_client.enums import ObjectGroupEnum, ObjectTypeEnum
from tionix_scheduler.core.helpers import get_version
from tionix_scheduler.tests.helpers import BaseAPITestCase
from tionix_scheduler.timetable.models import TimeTable


class TestTimeTable(BaseAPITestCase):
    """
    TestCase TimeTable CRUD.
    """
    @staticmethod
    def __fix_timetable(timetable):
        """
        Need to add id after creation.
        """
        timetable.update((('id', 1), ))

    def test_create_timetable_interval(self):
        """
        Ensure we can create a new timetable object with interval-schedule.
        """
        timetable, periodic_task, data = self.manager.get_interval_data()
        response = self.manager.create_timetable(data)

        # check response after post
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.content)

        # check timetable object
        self.__fix_timetable(timetable)
        data = response.data
        for k, v in timetable.iteritems():
            self.assertIn(k, data)
            self.assertEqual(v, data[k])

        # check periodic task object
        periodic_task_obj = TimeTable.objects.get(pk=1).periodic_task

        periodic_task_obj_params = (
            periodic_task_obj.task,
            periodic_task_obj.enabled,
            str(periodic_task_obj.interval),
            periodic_task_obj.args
        )
        self.assertEqual(
            (self.test_task_items[0].task,
             json.loads(periodic_task['activity']),
             'every 20 minutes',
             periodic_task['arguments']),
            periodic_task_obj_params)

    def test_create_timetable_interval_every(self):
        """
        Ensure we can't create a new timetable object with interval-schedule
        with empty interval every attribute.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        data['interval']['every'] = ''
        response = self.manager.create_timetable(data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_timetable_interval_period(self):
        """
        Ensure we can't create a new timetable object with interval-schedule
        with empty period attribute.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        data['interval']['period'] = ''
        response = self.manager.create_timetable(data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_timetable_interval_nonexistent_task(self):
        """
        Ensure we can't create a new timetable object with nonexistent task
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()

        data['unique_name'] = 'task does not exist'

        response = self.manager.create_timetable(data)

        self.assertEqual(
            response.data,
            u'Task not found - invalid unique_name'
        )

    def test_create_timetable_interval_no_task(self):
        """
        Ensure we can't create a new timetable object with nonexistent task
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()

        del data['unique_name']

        response = self.manager.create_timetable(data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_timetable_no_interval_no_crontab(self):
        """
        Ensure we can't create a new timetable object with nonexistent task
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()

        del data['interval']

        response = self.manager.create_timetable(data)

        self.assertEqual(
            response.data,
            u"[u'One of (interval, crontab) parameters required']"
        )

    def test_create_timetable_crontab(self):
        """
        Ensure we can create a new timetable object with crontab-schedule.
        """
        timetable, periodic_task, data = self.manager.get_crontab_data()
        response = self.manager.create_timetable(data)

        # check response after post
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.content)

        # check timetable object
        self.__fix_timetable(timetable)
        for k, v in timetable.iteritems():
            self.assertIn(k, response.data)
            self.assertEqual(v, response.data[k])

        # check periodic task object
        periodic_task_obj = TimeTable.objects.get(pk=1).periodic_task

        periodic_task_obj_params = (
            periodic_task_obj.task,
            periodic_task_obj.enabled,
            str(periodic_task_obj.crontab),
            periodic_task_obj.args
        )
        self.assertEqual(
            (self.test_task_items[0].task,
             json.loads(periodic_task['activity']),
             '* * * 01 01 (m/h/d/dM/MY)',
             periodic_task['arguments']),
            periodic_task_obj_params)

    def test_create_timetable_crontab_interval(self):
        """
        Ensure we can't create a new timetable object with crontab-schedule
        with interval not in range.
        """
        min_interval, max_interval = settings.TASK_INTERVAL_RANGE

        _timetable, _periodic_task, data = self.manager.get_crontab_data()

        data['crontab']['minute'] = '*/{0}'.format(max_interval + 1)

        response = self.manager.create_timetable(data)

        self.assertNotEqual(
            response.status_code, status.HTTP_201_CREATED,
            response.content)

        self.assertIn(
            u'interval must be in range from {0} to {1}'.format(min_interval,
                                                                max_interval),
            response.content)

        data['crontab']['minute'] = '*/test'

        response = self.manager.create_timetable(data)

        self.assertNotEqual(
            response.status_code, status.HTTP_201_CREATED,
            response.content)

        self.assertIn(u'interval must be integer', response.content)

        data['crontab']['minute'] = '*/{0}'.format(max_interval)

        response = self.manager.create_timetable(data)

        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED,
            response.content)

    def test_patch_timetable(self):
        """
        Ensure we can patch a new timetable object with schedule.
        """
        timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        patch = {'author_name': 'new admin'}
        timetable.update(patch)

        response = self.manager.patch_timetable(timetable, '1')

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.__fix_timetable(timetable)
        for k, v in timetable.iteritems():
            self.assertIn(k, response.data)
            self.assertEqual(v, response.data[k])

        activity = False
        patch = {'activity': activity}
        timetable.update(patch)
        response = self.manager.patch_timetable(timetable, '1')

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)
        periodic_task = response.data.get('periodic_task')
        self.assertEqual(periodic_task['enabled'], activity)

    def test_patch_timetable_expires(self):

        timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        _now = datetime.datetime.now()

        patch = {'expires': _now.isoformat()}
        timetable.update(patch)

        response = self.manager.patch_timetable(timetable, '1')

        parsed = dateutil.parser.parse(
            response.data['periodic_task']['expires'])
        self.assertIn(_now.isoformat(), parsed.isoformat())

    def test_patch_timetable_arguments(self):

        timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        new_arguments = ['new_user', 'new_password', 'new_tenant',
                         'new_subject_id', 'new_subject', 'new_user_domain']

        patch = {'arguments': json.dumps(new_arguments)}
        timetable.update(patch)

        response = self.manager.patch_timetable(timetable, '1')

        self.assertListEqual(json.loads(
            response.data['periodic_task']['args']), new_arguments)

    def test_patch_timetable_wrong_arguments(self):

        timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        new_arguments = json.dumps(['new_user', 'new_password'])

        patch = {'arguments': new_arguments}
        timetable.update(patch)

        response = self.manager.patch_timetable(timetable, '1')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn('Parameter "arguments" is not valid', response.data)

    def test_patch_timetable_begins(self):

        bad_msg = ('Parameters "begin_date" and "begin_time" '
                   'both need to be filled')

        timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)
        now_date = datetime.datetime.now().date()

        patch = {'begin_date': now_date.isoformat()}
        timetable.update(patch)

        response = self.manager.patch_timetable(timetable, '1')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn(bad_msg, response.data)

        timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)
        now_time = datetime.datetime.now().time()

        patch = {'begin_time': now_time.isoformat()}
        timetable.update(patch)

        response = self.manager.patch_timetable(timetable, '1')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn(bad_msg, response.data)

        patch = {'begin_date': now_date.isoformat()}
        timetable.update(patch)

        begins = datetime.datetime.combine(now_date, now_time)
        response = self.manager.patch_timetable(timetable, '1')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        kwargs = json.loads(response.data['periodic_task']['kwargs'])
        self.assertEqual(begins.isoformat(), kwargs['begins'])

    def test_put_timetable(self):
        """
        Ensure we can put a new timetable object with schedule.
        """
        timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        patch = {
            'interval': {
                'every': '30',
                'period': 'minutes'
            }
        }
        data.update(patch)

        response = self.manager.put_timetable(data, '1')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.__fix_timetable(timetable)
        for k, v in timetable.iteritems():
            self.assertIn(k, response.data)
            self.assertEqual(v, response.data[k])

        periodic_task_obj = TimeTable.objects.get(pk=1).periodic_task
        self.assertEqual(periodic_task_obj.interval.every, 30)

    def test_put_timetable_error(self):
        """
        Ensure we get Validation and NotFound error while put timetable object
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        interval = data.pop('interval')

        response = self.manager.put_timetable(data, '1')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST,
                         response.content)
        data['interval'] = interval

        data['unique_name'] = u'test'

        response = self.manager.put_timetable(data, '1')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND,
                         response.content)

    def test_delete_timetable(self):
        """
        Ensure we can delete a timetable object with schedule.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        response = self.manager.delete_timetable('1')

        self.assertEqual(
            response.status_code, status.HTTP_204_NO_CONTENT, response.content)

    def test_delete_subject_timetable(self):
        """
        Ensure we can delete objects by given subject_id and subject_type.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        subject_id = data['subject_id']
        subject_type = data['subject_type']

        response = self.manager.delete_subject_timetable(
            subject_id, subject_type)

        self.assertEqual(
            response.status_code, status.HTTP_200_OK)

        _filter_opts = {'subject_id': subject_id,
                        'subject_type': subject_type,
                        'deleted': 1}
        timetables = TimeTable.objects.filter(**_filter_opts)
        self.assertGreater(timetables.count, 0)

    def test_get_only_non_deleted_tasks(self):
        """
        Ensure we can get non executed tasks by given
        subject_id and subject_type.
        """
        _timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        response = self.manager.get_monitoring_tasks_list(
            {
                'subject_id': '2',
                'subject_type': ObjectTypeEnum.hypervisor,
                'subject_group': ObjectGroupEnum.admin,
                'show_deleted': False,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)
        content = json.loads(response.content)
        self.assertEqual(content['total_count'], 1)

        timetable = TimeTable.objects.all()[0]
        timetable.deleted = 1
        timetable.save()

        response = self.manager.get_monitoring_tasks_list(
            {
                'subject_id': '2',
                'subject_type': ObjectTypeEnum.hypervisor,
                'subject_group': ObjectGroupEnum.admin,
                'show_deleted': False,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)
        content = json.loads(response.content)
        self.assertEqual(content['total_count'], 0)

    def test_delete_subject_wrong_id(self):
        """
        Ensure API returns 404 if no items found to delete.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        subject_id = 'wrong_id'
        subject_type = data['subject_type']

        response = self.manager.delete_subject_timetable(
            subject_id, subject_type)

        self.assertEqual(
            response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_subject_no_params(self):
        """
        Ensure API returns 400 if not enough parameters are given.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        subject_id = data['subject_id']
        subject_type = data['subject_type']

        response = self.manager.delete_subject_timetable()
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST)

        response = self.manager.delete_subject_timetable(None, subject_type)
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST)

        response = self.manager.delete_subject_timetable(subject_id, None)
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_get_timetable_by_id(self):
        """
        Ensure we can get a timetable object with schedule by given id.
        """
        timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        response = self.manager.get_timetable('1')

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.__fix_timetable(timetable)
        for k, v in timetable.iteritems():
            self.assertIn(k, response.data)
            self.assertEqual(v, response.data[k])

    @mock.patch('tionix_scheduler.timetable.views.get_user_info',
                new=mock.Mock(return_value=('invalid_user',
                                            ['invalid_tenant'])))
    def test_get_timetable_by_id_forbidden(self):
        """
        Ensure we can't get a timetable object with schedule by given id with
        credentials that don't match with task author.
        """
        _timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        response = self.manager.get_timetable('1')

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_get_all_tasks(self):
        """
        Ensure we can get all available tasks.
        """
        response = self.manager.get_tasks_list(
            type='all', groups='all', author_name='admin', token='tokenuuid')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        task = self.test_task_items[0]
        self.assertIn(
            [task.unique_code, task.verbose_name],
            json.loads(response.content))

    def test_get_group_tasks(self):
        """
        Ensure we can get available tasks by group.
        """
        response = self.manager.get_tasks_list(type='all', groups='test_group')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.assertEqual(
            json.loads(response.content),
            [[task.unique_code, task.verbose_name]
             for task in self.test_task_items
             if 'test_group' in task.subject_groups])

    def test_get_type_tasks(self):
        """
        Ensure we can get available tasks by type.
        """
        response = self.manager.get_tasks_list(type='test_type', groups='all')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.assertEqual(
            json.loads(response.content),
            [[task.unique_code, task.verbose_name]
             for task in self.test_task_items
             if task.subject_type == 'test_type'])

    def test_get_group_tasks_multi_groups(self):
        """
        Ensure we can get available tasks by groups.
        """
        response = self.manager.get_tasks_list(
            type='all', groups=['test_group', 'test_group_2']
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.assertEqual(
            json.loads(response.content),
            [[task.unique_code, task.verbose_name]
             for task in self.test_task_items])

    def test_get_group_tasks_with_locale(self):
        """
        Ensure we can get tasks with passed locale.
        """
        response = self.manager.get_tasks_list(
            type='all',
            groups='test_group',
            locale='ru'
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.assertEqual(
            json.loads(response.content),
            [[task.unique_code, task.get_verbose_name('ru')]
             for task in self.test_task_items
             if 'test_group' in task.subject_groups])

    def test_get_group_tasks_no_group(self):
        """
        Ensure we get empty list if params without groups
        """
        response = self.manager.get_tasks_list()

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        self.assertEqual(json.loads(response.content), [])

    def test_get_group_tasks_no_existing_group(self):
        """
        Ensure we get empty list if there are no tasks with group
        """
        response = self.manager.get_tasks_list(groups='does not exist')

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        self.assertEqual(json.loads(response.content), [])

    def test_get_timetables(self):
        """
        Ensure we can get timetable records.
        """
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        self.manager.create_timetable(data)

        _timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.manager.create_timetable(data)

        response = self.manager.get_monitoring_tasks_list(
            {'tenant_id': self.test_tenant_id})
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        self.assertEqual(
            len(json.loads(response.content)),
            TimeTable.objects.count()
        )


class TestVersion(APITestCase):

    def setUp(self):
        self.client = APIClient()
        self.url = '/tnx-scheduler/'

    def test_version(self):

        response = self.client.get(self.url)

        info = {
            'info': {
                'version': get_version()
            }
        }

        response = json.loads(response.content)
        self.assertDictEqual(info, response)
