# coding: utf-8
# pylint: disable=no-member

import datetime
import json

from rest_framework import status
from tionix_scheduler_client.enums import ObjectTypeEnum, ObjectGroupEnum

from tionix_scheduler.tasks_controller import TaskController
from tionix_scheduler.timetable.models import TimeTable
from tionix_scheduler.tests.helpers import BaseAPITestCase


class TestViewTasksTimeTable(BaseAPITestCase):
    u""" testing view_tasks """

    def check_response(self, response):
        success_statuses = (status.HTTP_200_OK, status.HTTP_201_CREATED)
        self.assertIn(response.status_code, success_statuses, response.content)

    def setUp(self):
        super(TestViewTasksTimeTable, self).setUp()

        _timetable, _periodic_task, data = self.manager.get_interval_data()

        self.check_response(self.manager.create_timetable(data))
        self.check_response(self.manager.create_timetable(data))

        _timetable, _periodic_task, data = self.manager.get_crontab_data()
        self.check_response(self.manager.create_timetable(data))
        self.check_response(self.manager.create_timetable(data))
        self.check_response(self.manager.create_timetable(data))

    @property
    def interval_rows(self):
        return TimeTable.objects.exclude(
            periodic_task__interval__isnull=True).count()

    @property
    def crontab_rows(self):
        return TimeTable.objects.exclude(
            periodic_task__interval__isnull=False).count()

    @property
    def rows_created(self):
        return TimeTable.objects.count()

    def tearDown(self):
        super(TestViewTasksTimeTable, self).tearDown()
        TimeTable.objects.all().delete()

    def test_no_params(self):
        response = self.manager.get_monitoring_tasks_list(
            {'tenant_id': self.test_tenant_id})
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], self.rows_created)

        self.assertEqual(len(content['rows']), self.rows_created)

    def test_limit_offset(self):
        row_limit = 2
        response = self.manager.get_monitoring_tasks_list(
            {
                'offset': 0,
                'limit': row_limit,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], self.rows_created)
        self.assertEqual(len(content['rows']), row_limit)

        # Invalid offset/limit
        response = self.manager.get_monitoring_tasks_list(
            {'offset': 'test', 'limit': 'test',
             'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)
        # FIXME used wrong limit offset and returned all rows, should be 0
        self.assertEqual(len(content['rows']), self.rows_created)

        # only offset
        offset = 2
        response = self.manager.get_monitoring_tasks_list(
            {
                'offset': offset,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertEqual(len(content['rows']), self.rows_created - offset)

        # only limit
        response = self.manager.get_monitoring_tasks_list(
            {
                'limit': 2,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertEqual(len(content['rows']), 2)

    def test_disabled(self):
        _timetable, _periodic_task, data = self.manager.get_interval_data()
        data['activity'] = False

        self.check_response(self.manager.create_timetable(data))

        response = self.manager.get_monitoring_tasks_list(
            {'status': True, 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)
        # filter by enabled
        self.assertEqual(
            content['total_count'],
            TimeTable.objects.filter(periodic_task__enabled=True).count())

        response = self.manager.get_monitoring_tasks_list(
            {'status': False, 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)
        # expecting all rows
        self.assertEqual(content['total_count'], 1)

        # Invalid boolean
        response = self.manager.get_monitoring_tasks_list(
            {'status': 'test'}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 0)

    def test_show_deleted(self):

        self.manager.delete_timetable('1')
        self.manager.delete_timetable('2')

        response = self.manager.get_monitoring_tasks_list(
            {'show_deleted': True, 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)
        # filter by enabled
        self.assertEqual(
            content['total_count'],
            TimeTable.objects.all().count())

        response = self.manager.get_monitoring_tasks_list(
            {'show_deleted': False, 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)
        # expecting all rows
        self.assertEqual(
            content['total_count'],
            TimeTable.objects.filter(deleted=False).count())

    def test_by_subject_id(self):
        response = self.manager.get_monitoring_tasks_list(
            {
                'subject_id': '1',
                'subject_type': ObjectTypeEnum.instance,
                'subject_group': ObjectGroupEnum.admin,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 2)

        response = self.manager.get_monitoring_tasks_list(
            {'subject_id': '1'}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 0)

        response = self.manager.get_monitoring_tasks_list(
            {
                'subject_id': '10000',
                'subject_type': ObjectTypeEnum.instance,
                'subject_group': ObjectGroupEnum.admin,
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 0)

    def test_by_subject_type(self):
        response = self.manager.get_monitoring_tasks_list(
            {
                'subject_id': '1',
                'subject_type': ObjectTypeEnum.instance,
                'subject_group': ObjectGroupEnum.admin,
                'tenant_id': self.test_tenant_id
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 2)

        response = self.manager.get_monitoring_tasks_list(
            {'subject_id': '1', 'subject_type': 'does_not_exist'}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 0)

    def test_by_subject_group(self):

        response = self.manager.get_monitoring_tasks_list({
            'subject_type': ObjectTypeEnum.instance,
            'tenant_id': self.test_tenant_id
        })
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 2)

    def test_by_multi_subject_group(self):
        response = self.manager.get_monitoring_tasks_list(
            {'subject_type': [ObjectTypeEnum.instance,
                              ObjectTypeEnum.hypervisor],
             'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], self.rows_created)

    def test_filter(self):

        _timetable, _periodic_task, data = self.manager.get_interval_data()
        bad_rows = 0
        new_user_name = 'user'
        orig_user_name = data['author_name']
        unique_code2 = 'test_unique_code_2'
        unique_code3 = 'test_unique_code_3'
        data['unique_name'] = unique_code2
        data['author_name'] = new_user_name
        self.check_response(self.manager.create_timetable(data))
        bad_rows += 1
        data['repeat'] = 'false'
        data['unique_name'] = unique_code3
        data['author_name'] = orig_user_name
        self.check_response(self.manager.create_timetable(data))
        bad_rows += 1
        self.check_response(self.manager.create_timetable(data))
        bad_rows += 1

        # by repeat type
        response = self.manager.get_monitoring_tasks_list(
            {'type': 'true', 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], self.rows_created - bad_rows)

        response = self.manager.get_monitoring_tasks_list(
            {'type': 'false', 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 2)

        # by action name = unique name for scheduled task
        task_item = TaskController.get_task(unique_code2)

        self._get_user_info_patch(new_user_name)

        response = self.manager.get_monitoring_tasks_list({
            'action': task_item.get_verbose_name('en'),
            'author_name': new_user_name,
            'tenant_id': self.test_tenant_id,
        })

        self._get_user_info_patch(self.test_username)

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 1)

        task_item = TaskController.get_task(unique_code3)
        response = self.manager.get_monitoring_tasks_list({
            'action': task_item.get_verbose_name('en'),
            'tenant_id': self.test_tenant_id,
        })
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        # by author name
        username = 'admin'
        response = self.manager.get_monitoring_tasks_list(
            {
                'author_name': username,
                'tenant_id': self.test_tenant_id,
            }
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(
            content['total_count'],
            TimeTable.objects.filter(author_name=username).count())

        self._get_user_info_patch(new_user_name)

        username = 'user'
        response = self.manager.get_monitoring_tasks_list(
            {
                'author_name': username,
                'tenant_id': self.test_tenant_id,
            }
        )

        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(
            content['total_count'],
            TimeTable.objects.filter(author_name=username).count())

    def test_by_created_date(self):

        _timetable, _periodic_task, data = self.manager.get_interval_data()
        response = self.manager.get_monitoring_tasks_list({
            'author_name': data['author_name'],
            'tenant_id': data['tenant_id'],
            'start_date': datetime.date.today().strftime('%Y-%m-%d'),
            'end_date': (datetime.date.today() +
                         datetime.timedelta(days=1)).strftime('%Y-%m-%d'),

        })
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], self.rows_created)

    def test_by_created_date_no_tasks_on_date(self):
        response = self.manager.get_monitoring_tasks_list(
            {'start_date': '2015-01-01',
             'end_date': '2015-01-02'}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 0)

    def test_by_created_date_invalid_dates(self):
        response = self.manager.get_monitoring_tasks_list(
            {'start_date': '2015',
             'end_date': 'test'}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], 0)

    def test_sort_key_desc(self):
        response = self.manager.get_monitoring_tasks_list(
            {'sort_key': '-id',
             'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)
        rows = content['rows']
        self.assertTrue(bool(rows))
        self.assertEqual(rows[0]['id'], self.rows_created)

    def test_sort_key_asc(self):
        response = self.manager.get_monitoring_tasks_list(
            {'sort_key': 'id',
             'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)
        rows = content['rows']
        self.assertTrue(bool(rows))
        self.assertEqual(rows[0]['id'], 1)

    def test_locale(self):
        response = self.manager.get_monitoring_tasks_list(
            {'locale': 'ru', 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['rows'][0]['action'], u'Тестовое задание 1')

    def test_locale_does_not_exist(self):
        response = self.manager.get_monitoring_tasks_list(
            {'locale': 'asdf', 'tenant_id': self.test_tenant_id}
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content)

        content = json.loads(response.content)

        self.assertEqual(content['rows'][0]['action'], u'')

    def test_get_all_object_types(self):

        response = self.manager.get_tasks_objects()
        self.assertEqual(
            response.status_code, status.HTTP_200_OK)

        content = json.loads(response.content)
        object_types = content['object_types']
        for task_item in self.test_task_items:
            task_type = task_item.subject_type
            self.assertIn(task_type, object_types)
            for group in task_item.subject_groups:
                self.assertIn(group, object_types[task_type])

    def test_empty_tenant_id_param(self):

        self._get_user_info_patch(self.test_username, [])

        response = self.manager.get_monitoring_tasks_list()
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertGreater(self.rows_created, 0)

        self.assertEqual(content['total_count'], 0)

        self.assertEqual(len(content['rows']), 0)

    def test_unexists_field(self):

        field_name = 'test_filed_unexists'
        field_value = 'test_filed_unexists_value'

        response = self.manager.get_monitoring_tasks_list(
            {'tenant_id': self.test_tenant_id, field_name: field_value})
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.content
        )

        content = json.loads(response.content)

        self.assertEqual(content['total_count'], self.rows_created)

        self.assertEqual(len(content['rows']), self.rows_created)
