# coding: utf-8

"""
TimeTableSerializer testing
"""

import unittest
import mock

from rest_framework.exceptions import ValidationError, NotFound

from tionix_scheduler.timetable.serializers import (
    TimeTableSerializer, MIN_ARGS_COUNT)
from tionix_scheduler.timetable.serializers import (
    ObjectTypeEnum, ObjectGroupEnum)


class TnxTenantsControllerTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.serializer = TimeTableSerializer()
        cls._type_enum = ObjectTypeEnum.types
        ObjectTypeEnum.types += ('test_type', )
        cls._group_enum = ObjectGroupEnum.groups
        ObjectGroupEnum.groups += ('test_group', )

    @classmethod
    def tearDownClass(cls):
        ObjectTypeEnum.types = cls._type_enum
        ObjectGroupEnum.groups = cls._group_enum

    def test_check_not_valid_param(self):
        self.assertRaises(ValidationError,
                          self.serializer.check_parameter,
                          None,
                          u'test')

    def test_check_not_valid_subj_type(self):
        subject = u'Test'
        subject_id = u'123'
        subject_type = u'test_type_invalid'
        subject_group = u'test_group'
        with self.assertRaises(NotFound) as ex:
            self.serializer.check_subject_info(subject, subject_id,
                                               subject_type, subject_group)
        self.assertEqual(
            unicode(ex.exception),
            u'Not found object type with name "{0}"'.format(subject_type))

    def test_check_not_valid_subj_group(self):
        subject = u'Test'
        subject_id = u'123'
        subject_type = u'test_type'
        subject_group = u'test_group_invalid'
        with self.assertRaises(NotFound) as ex:
            self.serializer.check_subject_info(subject, subject_id,
                                               subject_type, subject_group)
        self.assertEqual(
            unicode(ex.exception),
            u'Not found object group with name "{0}"'.format(subject_group))

    @mock.patch('tionix_scheduler.timetable.serializers.get_keystone_client',
                new=mock.Mock(side_effect=Exception()))
    def test_check_auth_info_not_valid(self):
        with self.assertRaises(ValidationError) as ex:
            self.serializer.check_auth_info(u'username', u'pass',
                                            u'tenant_id', u'test_user_domain')
        self.assertEqual(unicode(ex.exception),
                         u'[u\'Parameter "arguments" is not valid. \']')

    def test_check_arguments_not_valid(self):
        arguments = range(0, MIN_ARGS_COUNT - 1)
        self.assertRaises(ValidationError,
                          self.serializer.check_arguments,
                          unicode(arguments))

    @mock.patch(('tionix_scheduler.timetable.serializers.TimeTableSerializer.'
                 'check_unique_task_name'), new=mock.Mock())
    @mock.patch(('tionix_scheduler.timetable.serializers.TimeTableSerializer.'
                 'check_auth_info'), new=mock.Mock())
    @mock.patch(('tionix_scheduler.timetable.serializers.TimeTableSerializer.'
                 'check_subject_info'), new=mock.Mock())
    def test_check_data_not_valid(self):
        subject_id = u'4'
        subject = u'test'
        data = {
            'arguments': u'[1, 2, 3, "' + subject_id + '", "test1", "test"]',
            'interval': u'test',
            'subject_id': u'5',
            'subject': subject,
        }
        self.assertRaises(ValidationError,
                          self.serializer.check_data,
                          data)
        with self.assertRaises(ValidationError) as ex:
            self.serializer.check_data(data)

        self.assertEqual(unicode(ex.exception),
                         u'[u\'4 element in "arguments" parameter must be '
                         u'equal with "subject_id" parameter value "5"\']')

        data['subject_id'] = subject_id
        with self.assertRaises(ValidationError) as ex:
            self.serializer.check_data(data)

        self.assertEqual(unicode(ex.exception),
                         u'[u\'5 element in "arguments" parameter must be '
                         u'equal with "subject" parameter '
                         u'value "{0}"\']'.format(subject))

    def test_validate_interval(self):
        data = {}
        result = self.serializer.validate_interval(data)
        self.assertIsNone(result)

    def test_validate_crontab(self):
        data = {}
        result = self.serializer.validate_crontab(data)
        self.assertIsNone(result)
