import unittest
from collections import defaultdict

from tionix_scheduler.settings import (
    module_config_path, update_module_settings, update_global_from_settings,
    DATABASES
)
from tionix_scheduler.settings.settings_handler import (
    merge_dicts, get_valid_db_engine, config_sections)


class TestSettingHandler(unittest.TestCase):
    def test_merge_dicts(self):
        base_dict = {}
        update_dict = {}

        self.assertEqual(merge_dicts(base_dict, update_dict), base_dict)

        base_dict = {'key_1': {'key_2': 'value_2'}}
        update_dict = {}

        self.assertEqual(merge_dicts(base_dict, update_dict), base_dict)

        base_dict = {}
        update_dict = {'key_1': {'key_2': 'value_2'}}

        self.assertEqual(merge_dicts(base_dict, update_dict), update_dict)

        base_dict = {'key_1': '1'}
        update_dict = {'key_1': {'key_2': 'value_2'}}

        self.assertEqual(merge_dicts(base_dict, update_dict), update_dict)

        base_dict = {'key_1': {'key_2': 'value_2'}}
        update_dict = {'key_1': 'value_1'}

        self.assertEqual(merge_dicts(base_dict, update_dict), update_dict)

        base_dict = {'key_2': 'value_2'}
        update_dict = {'key_1': {'key_2': 'value_2'}}
        after_merge_dict = {'key_1': {'key_2': 'value_2'}, 'key_2': 'value_2'}

        self.assertEqual(merge_dicts(base_dict, update_dict), after_merge_dict)

        base_dict = {'key_1': {'key_1': 'value_1'}}
        update_dict = {'key_1': {'key_2': 'value_2'}}
        after_merge_dict = {'key_1': {'key_1': 'value_1', 'key_2': 'value_2'}}

        self.assertEqual(merge_dicts(base_dict, update_dict), after_merge_dict)

        base_dict = {
            'key_1': {'key_2': {'key_3': 'value_3', 'key_4': 'value_4'}}}
        update_dict = {
            'key_1': {'key_2': {'key_3': 'new_value_3'}, 'key_3': 'value_3'}}
        after_merge_dict = {
            'key_1': {'key_2': {'key_3': 'new_value_3', 'key_4': 'value_4'},
                      'key_3': 'value_3'}}

        self.assertEqual(merge_dicts(base_dict, update_dict), after_merge_dict)

    def test_get_valid_db_engine(self):
        default_value = DATABASES['default']['ENGINE']

        test_data = (
            ('mysql', 'django.db.backends.mysql'),
            ('sqlite3', 'django.db.backends.sqlite3'),
            ('oracle', 'django.db.backends.oracle'),
            ('django.db.backends.sqlite3', 'django.db.backends.sqlite3'),
            ('invalid_engine', default_value),
            ('django.db.backends.invalid_engine', default_value),
        )

        for test_value, expected_value in test_data:
            validated = get_valid_db_engine('key', test_value, default_value)
            self.assertEqual(validated, expected_value)

    def test_update_global_from_settings(self):

        def get_example_file_path(cfg_path):
            return cfg_path[1:] + '.example'

        module_cfg_path = get_example_file_path(module_config_path)

        test_globals = {
            'DATABASES': {
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': 'tionix_scheduler',
                },
            },
            'KEYSTONE': {
                'auth_user': 'admin',
                'auth_password': 'admin',
                'auth_tenant': 'admin',
                'auth_url': 'http://localhost:5000',
                'auth_version': '3',
                'compute_service_name': 'compute',
                'volume_service_name': 'volumev2',
                'user_domain_name': 'default',
                'project_domain_name': 'default'
            },
            'RABBIT_QUEUES': {
                'username': 'tionix',
                'password': 'password',
                'host': 'localhost',
                'broker_type': 'amqp',
                'port': '5672'
            },
            'LOG_LEVEL': 'INFO',
        }

        module_settings = defaultdict(dict)
        update_module_settings(module_settings, test_globals, module_cfg_path,
                               '')
        update_global_from_settings(module_settings, test_globals)

        for key, value in module_settings.items():
            if isinstance(value, dict):
                for k, v in value.items():
                    self.assertEqual(v, test_globals[key][k])
            else:
                self.assertEqual(value, test_globals[key])

        for section in config_sections:
            module_settings[section] = None

        update_global_from_settings(module_settings, test_globals)
        for section in config_sections:
            if section == 'DB':
                self.assertIsNotNone(test_globals['DATABASES']['default'])
            else:
                self.assertIsNotNone(test_globals[section])
