# coding: utf-8
# pylint: disable=no-member

import datetime
import json

from json import dumps, loads

from django.conf import settings
from django.db import transaction

from djcelery.models import (
    IntervalSchedule, CrontabSchedule, PERIOD_CHOICES)

from rest_framework import serializers
from rest_framework.exceptions import ValidationError, NotFound
from tionix_client.os_clients.keystone_client import get_keystone_client

from tionix_scheduler.tasks_controller import TaskController
from tionix_scheduler.timetable.models import TimeTable, TnxPeriodicTask

from tionix_scheduler_client.enums import (
    ObjectTypeEnum, ObjectGroupEnum)

# min_args_count - minimum arguments to create task, others are
# optional for each task
MIN_ARGS_COUNT = 6


class IntervalScheduleSerializer(serializers.Serializer):

    every = serializers.IntegerField()
    period = serializers.ChoiceField(choices=PERIOD_CHOICES)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass


class CrontabScheduleSerializer(serializers.Serializer):
    minute = serializers.CharField(
        max_length=64, default='*', allow_blank=True)
    hour = serializers.CharField(
        max_length=64, default='*', allow_blank=True)
    day_of_week = serializers.CharField(
        max_length=64, default='*', allow_blank=True)
    day_of_month = serializers.CharField(
        max_length=64, default='*', allow_blank=True)
    month_of_year = serializers.CharField(
        max_length=64, default='*', allow_blank=True)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass


class PeriodicTaskSerializer(serializers.ModelSerializer):

    interval = IntervalScheduleSerializer()
    crontab = CrontabScheduleSerializer()

    next_run_time = serializers.SerializerMethodField()

    class Meta:  # pylint: disable=W0232,C1001
        model = TnxPeriodicTask
        fields = '__all__'

    @staticmethod
    def get_next_run_time(obj):
        if obj.enabled:
            now = datetime.datetime.now()
            last_run_at = obj.last_run_at or now
            next_run_time = (
                now + obj.schedule.remaining_estimate(last_run_at) +
                datetime.timedelta(seconds=1))
        else:
            next_run_time = None
        return next_run_time


class TimeTableSerializer(serializers.ModelSerializer):

    author_name = serializers.CharField(required=True)
    tenant_id = serializers.CharField(required=True)

    activity = serializers.BooleanField(required=False, initial='true')
    expires = serializers.DateTimeField(required=False, allow_null=True)

    periodic_task = PeriodicTaskSerializer(required=False)

    interval = IntervalScheduleSerializer(required=False, allow_null=True)
    crontab = CrontabScheduleSerializer(required=False, allow_null=True)

    # JSON encoded positional arguments
    arguments = serializers.CharField(required=False, initial='[]')

    # JSON encoded positional kwargs
    key_arguments = serializers.CharField(required=False, initial='{}')

    run_count_limit = serializers.IntegerField(required=False, allow_null=True)

    interval_param = u'interval'
    crontab_param = u'crontab'
    subject_param = u'subject'
    subject_id_param = u'subject_id'
    subject_type_param = u'subject_type'
    subject_group_param = u'subject_group'
    arguments_param = u'arguments'
    unique_name_param = u'unique_name'

    class Meta:  # pylint: disable=W0232,C1001
        model = TimeTable
        fields = ('id', 'author_name', 'tenant_id', 'description',
                  'unique_name', 'repeat', 'subject', 'begin_date',
                  'begin_time', 'timezone', 'activity', 'expires', 'interval',
                  'crontab', 'arguments', 'created', 'periodic_task',
                  'subject_type', 'subject_group', 'subject_id',
                  'run_count_limit', 'key_arguments', 'last_run_message',
                  'last_run_result')

        extra_kwargs = {
            'token': {'allow_blank': True},
        }

        depth = 2

        read_only_fields = ('created', 'periodic_task', 'last_run_message',
                            'last_run_result')

    @staticmethod
    def validate_interval(data):
        """
        Makes data valid if no interval.
        """
        if not data.get('every') and not data.get('period'):
            data = None
        return data

    @staticmethod
    def validate_crontab(data):
        """
        Makes data valid if no crontab.
        """
        crontab_list = [
            data[i] for i in ('minute', 'hour', 'day_of_week',
                              'day_of_month', 'month_of_year')
            if data.get(i)
        ]
        if not crontab_list:
            data = None
        else:
            for i in crontab_list:
                if i and '/' in i:
                    interval = i.rsplit('/', 1)[-1]
                    if not interval.isdigit():
                        raise ValidationError(u'interval must be integer')
                    min_interval, max_interval = settings.TASK_INTERVAL_RANGE
                    if not min_interval <= int(interval) <= max_interval:
                        raise ValidationError(
                            u'interval must be in range '
                            u'from {0} to {1}'.format(
                                min_interval, max_interval))

        return data

    def __build_schedule(self, validated_data):
        """
        Makes interval or crontab.
        """
        schedule = None

        # create schedule
        if self.interval_param in validated_data:
            interval = validated_data.pop(self.interval_param)
            if interval:
                schedule, _created = IntervalSchedule.objects.get_or_create(
                    **interval)
        if self.crontab_param in validated_data:
            crontab = validated_data.pop(self.crontab_param)
            if crontab:
                schedule, _created = CrontabSchedule.objects.get_or_create(
                    **crontab)
        return schedule

    def __new_task(self, validated_data):
        """
        Creates new periodic task record.
        """
        self.check_data(validated_data)

        # find task-item
        task_item = TaskController.get_task(validated_data.get(
            self.unique_name_param))

        # extra options to load task like queue etc.
        options = {}
        # activity
        if 'activity' in validated_data:
            options['enabled'] = validated_data.pop('activity')

        # args to load task
        options['args'] = validated_data.pop('arguments')

        kwargs = validated_data.pop('key_arguments', {})
        if kwargs:
            try:
                kwargs = loads(kwargs)
            except ValueError:
                kwargs = {}

        if validated_data['repeat']:
            begin_date = validated_data['begin_date']
            begin_time = validated_data['begin_time']

            if begin_date and begin_time:
                begins = datetime.datetime.combine(begin_date, begin_time)
                kwargs['begins'] = begins.isoformat()

        if 'expires' in validated_data:
            options['expires'] = validated_data.pop('expires')

        # create task
        periodic_task = task_item.apply_schedule(
            self.__build_schedule(validated_data), options, **kwargs)

        return periodic_task

    @transaction.atomic
    def create(self, validated_data):
        # create task
        periodic_task = self.__new_task(validated_data)

        # create timetable
        timetable = TimeTable.objects.create(
            periodic_task=periodic_task, **validated_data)

        # pass timetable_id into kwargs
        if isinstance(periodic_task.kwargs, basestring):
            kwargs = json.loads(periodic_task.kwargs)
        else:
            kwargs = periodic_task.kwargs

        kwargs.update(timetable_id=str(timetable.pk))
        periodic_task.kwargs = json.dumps(kwargs)
        periodic_task.save()

        return timetable

    @transaction.atomic
    def update(self, instance, validated_data):

        # only non-partial update can remove task
        if not self.partial:
            # PUT-request
            instance.periodic_task.delete()
            instance.periodic_task = self.__new_task(validated_data)

        else:
            # PATCH-request
            updatable_fields = ['expires']
            periodic_task_updated = False

            arguments = validated_data.pop('arguments', None)

            # change task call arguments
            if arguments:
                args = loads(arguments)
                old_args = loads(instance.periodic_task.args)

                if len(args) < 3:
                    raise ValidationError(u'Parameter "{0}" is not valid.'
                                          u''.format(self.arguments_param))

                args_changed = any(a != b for a, b in zip(args, old_args))

                if len(args) != len(old_args) or args_changed:

                    (username, password, tenant_id, _subject_id_arg,
                     _subject_name_arg,
                     user_domain_name) = args[:MIN_ARGS_COUNT]

                    self.check_auth_info(username, password, tenant_id,
                                         user_domain_name)
                    instance.periodic_task.args = arguments
                    periodic_task_updated = True

            # change other updatable periodic task fields
            for field in updatable_fields:
                old_value = getattr(instance.periodic_task, field)
                value = validated_data.get(field, old_value)
                if value != old_value:
                    setattr(instance.periodic_task, field, value)
                    periodic_task_updated = True

            subject = validated_data.get(
                self.subject_param) or instance.subject
            subject_id = validated_data.get(
                self.subject_id_param) or instance.subject_id
            subject_type = validated_data.get(
                self.subject_type_param) or instance.subject_type
            subject_group = validated_data.get(
                self.subject_group_param) or instance.subject_group
            self.check_subject_info(subject, subject_id, subject_type,
                                    subject_group)

            # enable/disable schedule
            if 'activity' in validated_data:
                instance.periodic_task.enabled = validated_data.pop('activity')
                periodic_task_updated = True

            # change begins field
            kwargs = loads(instance.periodic_task.kwargs)
            begin_date = validated_data.pop('begin_date', None)
            begin_time = validated_data.pop('begin_time', None)
            if begin_date and begin_time:
                begins = datetime.datetime.combine(begin_date, begin_time)
                kwargs['begins'] = begins.isoformat()
                instance.periodic_task.kwargs = dumps(kwargs)
                periodic_task_updated = True
            elif begin_date or begin_time:
                raise ValidationError(u'Parameters "begin_date" and '
                                      u'"begin_time" both need to be filled.')

            if periodic_task_updated:
                instance.periodic_task.save()

        for attr, value in validated_data.iteritems():
            if hasattr(instance, attr):
                attr_value = getattr(instance, attr)
                if value != attr_value:
                    setattr(instance, attr, value)

        instance.save()

        return instance

    @staticmethod
    def check_parameter(parameter, parameter_name):
        """
        Raise error if parameter is None
        """
        if parameter is None:
            raise ValidationError(u'Parameter "{0}" required'.format(
                parameter_name))

    def check_subject_info(self, subject, subject_id, subject_type,
                           subject_group):
        """
        Check subject info (subject name, subject id, subject type)
        """
        self.check_parameter(subject, self.subject_param)
        self.check_parameter(subject_id, self.subject_id_param)
        self.check_parameter(subject_type, self.subject_type_param)
        self.check_parameter(subject_group, self.subject_group_param)

        if subject_type not in ObjectTypeEnum.types:
            msg = (u'Not found object type with '
                   u'name "{0}"'.format(subject_type))
            raise NotFound(msg)

        if subject_group not in ObjectGroupEnum.groups:
            msg = (u'Not found object group with '
                   u'name "{0}"'.format(subject_group))
            raise NotFound(msg)

    def check_auth_info(self, username, password, tenant_id, user_domain_name):
        """
        Check authentication info sending in request
        """
        try:
            client = get_keystone_client(username=username,
                                         password=password,
                                         tenant_id=tenant_id,
                                         user_domain_name=user_domain_name)
        except Exception as err:
            raise ValidationError(u'Parameter "{0}" is not valid. '
                                  u'{1}'.format(self.arguments_param, err))
        return client

    def check_arguments(self, arguments):
        """
        Check arguments param
        """
        self.check_parameter(arguments, self.arguments_param)
        list_arguments = loads(arguments)
        if len(list_arguments) < MIN_ARGS_COUNT:
            msg = (u'Parameter "{0}" must be a list with a MINIMUM of '
                   u'{1} elements such as: [<user name>, <password>, '
                   u'<tenant id>, <subject id>, <subject name>, '
                   u'<user_domain_name>].'
                   .format(self.arguments_param, MIN_ARGS_COUNT))

            raise ValidationError(msg)
        return list_arguments

    def check_unique_task_name(self, unique_name):
        """
        Check unique_name param
        """
        self.check_parameter(unique_name, self.unique_name_param)
        task_item = TaskController.get_task(unique_name)
        if not task_item:
            raise NotFound(u'Task not found - invalid '
                           u'{0}'.format(self.unique_name_param))

    def check_data(self, data):
        """
        Check request params
        """
        subject = data.get(self.subject_param)
        subject_id = data.get(self.subject_id_param)
        subject_type = data.get(self.subject_type_param)
        subject_group = data.get(self.subject_group_param)
        unique_name = data.get(self.unique_name_param)
        arguments = data.get(self.arguments_param)

        self.check_unique_task_name(unique_name)
        list_arguments = self.check_arguments(arguments)

        ((username, password, tenant_id, subject_id_arg, subject_name_arg,
          user_domain_name), _others) = (list_arguments[:MIN_ARGS_COUNT],
                                         list_arguments[MIN_ARGS_COUNT:])

        self.check_auth_info(username, password, tenant_id, user_domain_name)
        self.check_subject_info(subject, subject_id, subject_type,
                                subject_group)

        if unicode(subject_id) != unicode(subject_id_arg):
            msg = (u'4 element in "{0}" parameter must be equal '
                   u'with "{1}" parameter value "{2}"'
                   .format(self.arguments_param, self.subject_id_param,
                           subject_id))

            raise ValidationError(msg)

        if subject != subject_name_arg:
            msg = (u'5 element in "{0}" parameter must be equal '
                   u'with "{1}" parameter value "{2}"'
                   .format(self.arguments_param, self.subject_param, subject))

            raise ValidationError(msg)

        if (not data.get(self.interval_param) and
                not data.get(self.crontab_param)):
            raise ValidationError(
                u'One of ({0}, {1}) parameters required'.format(
                    self.interval_param, self.crontab_param))
