# coding: utf-8
# pylint: disable=no-member, too-many-statements, too-many-locals

from collections import defaultdict
from datetime import datetime
import logging

from django.db import transaction

from rest_framework import status, viewsets
from rest_framework.exceptions import ValidationError, NotFound
from rest_framework.decorators import list_route
from rest_framework.response import Response

from tionix_scheduler.tasks_controller import TaskController
from tionix_scheduler.timetable.helpers import get_user_info
from tionix_scheduler.timetable.models import TimeTable
from tionix_scheduler.timetable.serializers import TimeTableSerializer
from tionix_scheduler.timetable.service import TasksActions, NoTasksFoundError

logger = logging.getLogger(__name__)


class TimeTableViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows users to be viewed or edited.
    """
    queryset = TimeTable.objects.all()
    serializer_class = TimeTableSerializer

    DEFAULT_LIMIT = 10

    def get_queryset(self):
        return self.queryset.select_related('periodic_task')

    def perform_destroy(self, instance):
        """
        :type instance: tionix_scheduler.timetable.models.TimeTable
        """

        TasksActions.delete_timetable(instance)

    @transaction.atomic
    def destroy(self, request, *args, **kwargs):
        return super(TimeTableViewSet, self).destroy(request, *args, **kwargs)

    @list_route()
    def tasks_list(self, request):
        """
        Available tasks to schedule provided by python packages
        collected by TaskController.
        query_params:
            groups: list of groups;
                    filter by task group - belonging of task-module
            locale: string; eg.: ru
                    Result task info by chosen locale
        """
        tasks = TaskController.task_collection

        filtered_tasks = []

        subject_type = request.query_params.get('type')
        groups = request.query_params.getlist('groups')
        locale = request.query_params.get('locale')

        for _e_point, tasks in tasks.iteritems():
            for unique_code, task in tasks.iteritems():
                if subject_type == 'all' or task.subject_type == subject_type:
                    if ('all' in groups or any(g for g in task.subject_groups
                                               if g in groups)):
                        filtered_tasks.append(
                            (unique_code, task.get_verbose_name(locale)))

        return Response(filtered_tasks)

    @list_route(methods=['GET', 'DELETE'])
    def tasks_monitoring(self, request):
        """
        Tasks from celery - scheduled tasks for object
        """
        result = None
        if request.method == 'DELETE':
            result = self.delete_tasks(request)
        else:
            result = self.view_tasks(request)
        return result

    @staticmethod
    def view_tasks(request):  # pylint: disable=too-many-locals

        def _date_is_valid(date):
            try:
                datetime.strptime(date, '%Y-%m-%d')
            except ValueError:
                return False
            return True

        def _get_param(param_name, param_value):
            return (
                request.query_params.get(param_name)
                if len(param_value) == 1
                else request.query_params.getlist(param_name)
            )

        def _bool_value(value, default=None):
            val_map = {
                'true': True,
                'false': False,
            }
            return val_map.get(value.lower(), default)

        def _task_value(value):
            codes = []
            all_tasks = TaskController.task_collection
            for _entry_point, task_subcollection in all_tasks.iteritems():
                for code, item in task_subcollection.iteritems():
                    verbose_name = item.get_verbose_name(locale).lower()
                    if value.lower() in verbose_name:
                        codes.append(code)
            return codes

        def _get_list(value):
            if isinstance(value, basestring):
                value = [value]
            return value

        def _get_ordered_tasks(tasks_sort_key, tasks_list):
            if isinstance(tasks_sort_key, basestring):
                sort_dir = ''
                if tasks_sort_key.startswith('-'):
                    _, sort_dir, tasks_sort_key = tasks_sort_key.partition('-')
                tasks_sort_key = ('{0}{1}'.format(
                    sort_dir, sort_map.get(tasks_sort_key, tasks_sort_key)),)
            return tasks_list.order_by(*tasks_sort_key)

        def _date_range(date_range):

            start_date, end_date = date_range

            if not start_date or not _date_is_valid(start_date):
                start_date = datetime.min

            if not end_date or not _date_is_valid(end_date):
                end_date = datetime.max

            return start_date, end_date

        query_params = {
            k: _get_param(k, v)
            for k, v in request.query_params.iterlists()
        }

        mandatory_query_params = {
            'tenant_id',
        }

        if not mandatory_query_params.issubset(set(query_params)):
            return Response({
                'rows': [],
                'total_count': 0
            })

        model_fields = TimeTable._meta.get_all_field_names()

        sort_key = query_params.pop('sort_key', [])
        locale = query_params.pop('locale', '')

        username, tenants = get_user_info(request,
                                          query_params.get('tenant_id'))
        query_params['tenant_id'] = tenants
        query_params['author_name'] = username

        query_params['created'] = (query_params.pop('start_date', ''),
                                   query_params.pop('end_date', ''))

        offset = query_params.pop('offset', 0)
        if isinstance(offset, basestring):
            offset = int(offset) if offset.isdigit() else 0

        limit = query_params.pop('limit', TimeTableViewSet.DEFAULT_LIMIT)
        if isinstance(limit, basestring):
            limit = (int(limit)
                     if limit.isdigit()
                     else TimeTableViewSet.DEFAULT_LIMIT)

        show_deleted = _bool_value(
            query_params.pop('show_deleted', ''), default=False)

        skipped_query_params = (
            'all_tenants',
            'filter_tenants'
        )
        filter_map = {
            'name': ('description', '__icontains', None),
            'status': ('periodic_task', '__enabled', _bool_value),
            'type': ('repeat', '', _bool_value),
            'action': ('unique_name', '__in', _task_value),
            'author_name': ('author_name', '', None),
            'subject_id': ('subject_id', '__in', _get_list),
            'subject_type': ('subject_type', '__in', _get_list),
            'subject_group': ('subject_group', '__in', _get_list),
            'tenant_id': ('tenant_id', '__in', None),
            'created': ('created', '__range', _date_range),
        }

        sort_map = {
            'name': 'description',
            'status': 'periodic_task__enabled',
            'type': 'repeat',
            'action': 'unique_name',
        }

        _filter = {}
        for key, val in query_params.iteritems():
            if key in skipped_query_params:
                continue

            field = key
            if key in filter_map:
                field, suffix, modifier = filter_map[key]
                if callable(modifier):
                    val = modifier(val)
            elif isinstance(val, (list, tuple)):
                suffix = '__in'
            else:
                suffix = '__icontains'
            if field in model_fields:
                _filter[field + suffix] = val

        if not show_deleted:
            _filter['deleted'] = False

        tasks = TimeTable.objects.filter(**_filter).select_related(
            'periodic_task',
            'periodic_task__interval',
            'periodic_task__crontab'
        )
        total_count = tasks.count()

        if sort_key:
            tasks = _get_ordered_tasks(sort_key, tasks)

        if limit:
            tasks = tasks[offset:offset + limit]

        formatted_tasks = []

        for task in tasks.iterator():
            task_item = TaskController.get_task(task.unique_name)

            formatted_tasks.append({
                'id': task.id,
                'name': task.description,
                'unique_name': task_item.unique_code if task_item else '',
                'action': (
                    task_item.get_verbose_name(locale)
                    if task_item else 'task item not found'),
                'author_name': task.author_name,
                'tenant_id': task.tenant_id,
                'created': task.created,
                'expires': task.periodic_task.expires,
                'last_run_at': task.periodic_task.last_run_at,
                'schedule': str(task.periodic_task.schedule),
                'repeat': task.repeat,
                'subject': task.subject,
                'subject_type': task.subject_type,
                'subject_group': task.subject_group,
                'subject_id': task.subject_id,
                'enabled': task.periodic_task.enabled,
                'begin_date': task.begin_date,
                'begin_time': task.begin_time,
                'timezone': task.timezone,
                'total_run_count': task.periodic_task.total_run_count,
                'last_run_result': task.last_run_result,
                'deleted': task.deleted,
                'deleted_at': task.deleted_at
            })

        return Response({
            'rows': formatted_tasks,
            # Total filtered tasks count
            'total_count': total_count
        })

    @staticmethod
    def delete_tasks(request):
        subject_id = request.data.get('subject_id')
        subject_type = request.data.get('subject_type')

        _filter_opts = {'subject_id': subject_id,
                        'subject_type': subject_type}

        msg = u'Tasks has been successfully deleted'
        return_code = status.HTTP_200_OK

        try:
            TasksActions.delete_tasks_by_filter(**_filter_opts)
        except ValueError as e:
            msg = unicode(e)
            return_code = status.HTTP_400_BAD_REQUEST
        except NoTasksFoundError as e:
            msg = unicode(e)
            return_code = status.HTTP_404_NOT_FOUND
        except Exception as e:
            msg = unicode(e)
            return_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        return Response(data=msg, status=return_code)

    @list_route()
    def task_objects(self, request):
        """
        List of object types available for schedule.
        """
        group_tasks = TaskController.task_collection.items()
        object_types = defaultdict(set)
        for _group, tasks in group_tasks:
            for _task_name, task in tasks.items():
                groups = ([task.subject_groups]
                          if isinstance(task.subject_groups, basestring)
                          else task.subject_groups)
                object_types[task.subject_type].update(groups)

        response = {'object_types': object_types}

        return Response(response)

    def list(self, request, *args, **kwargs):
        return self.view_tasks(request)

    def create(self, request, *args, **kwargs):
        """
        Create new task
        """
        try:
            response = super(TimeTableViewSet, self).create(
                request, *args, **kwargs)
        except ValidationError as err:
            response = Response(data=unicode(err),
                                status=status.HTTP_400_BAD_REQUEST)
        except NotFound as err:
            response = Response(data=unicode(err),
                                status=status.HTTP_404_NOT_FOUND)
        except Exception as err:
            response = Response(data=unicode(err),
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        return response

    def update(self, request, *args, **kwargs):
        """
        Update task
        """
        try:
            response = super(TimeTableViewSet, self).update(
                request, *args, **kwargs)
        except ValidationError as err:
            response = Response(data=unicode(err),
                                status=status.HTTP_400_BAD_REQUEST)
        except NotFound as err:
            response = Response(data=unicode(err),
                                status=status.HTTP_404_NOT_FOUND)
        except Exception as err:
            response = Response(data=unicode(err),
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        return response

    def retrieve(self, request, *args, **kwargs):
        """
        Get single task
        """
        user_name, tenants = get_user_info(request)
        response = super(TimeTableViewSet, self).retrieve(request, *args,
                                                          **kwargs)
        author_name = response.data.get('author_name')
        tenant_id = response.data.get('tenant_id')

        if user_name != author_name or tenant_id not in tenants:
            msg = (u'Task with id "{0}" is not available for given '
                   u'credentials.'.format(kwargs.get('pk')))
            response = Response(data=msg, status=status.HTTP_403_FORBIDDEN)

        return response
